#!/usr/bin/perl -w
#
# $Id: man.cgi,v 1.3 1997/08/23 10:30:25 richard Exp $
#
#    man.cgi - view man pages over the web
#    Copyright (C) 1997 Richard Kettlewell
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Contact the author by email at <richard@greenend.org.uk>
#

use CGI;

$version = "1.0";

%sets = ("Linux" => "",
	 "UNIX V7" => "-M /usr/local/share/v7/man",
	 "BSD" => "-M /usr/local/share/ucb/man");
$q = new CGI;
if(!defined($command = $q->param("command"))) {
    $command = "";
}
if($command eq "") {
    print $q->header("text/html");
    print $q->start_html(-title=>"Man by web");
    print "<center><h1>CGI Man Pager Browser ($version)</h1></center><hr>\n";
    print $q->startform;
    print "Type of query:\n";
    print $q->popup_menu(-name=>"command",
			 -values=>["man","apropos","whatis"]);
    print "<p>\n";
    print "Search string:\n";
    print $q->textfield(-name=>"page",
			-default=>"");
    print "<p>\n";
    print "Man page set:\n";
    print $q->popup_menu(-name=>"set",
			 -values=>[keys %sets]);
    print "<p>\n";
    print $q->submit(-name=>"search");
    print $q->endform;
    print "<hr>\n";
    print "(c) 1996,1997 <a href=\"mailto:richard\@greenend.org.uk\">Richard Kettlewell</a>.  Subject to GNU GPL.<p>\n";
    print $q->end_html;
    exit 0;
}
$page = $q->param("page");
if($page =~ /\((\d)\w*\)$/) {
    $section = $1;
    $page =~ s/\s*\((\d\w*)\)$//;
}
$set = $q->param("set");
if(exists $sets{$set}) {
    $M = $sets{$set};
} else {
    $M = "";
}
if($page !~ /^[a-zA-Z0-9_\-\.:]+$/) {
    &fail("Invalid page");
}
if(defined $section && $section !~ /^[a-zA-Z0-9_\-\.]+$/) {
    &fail("Invalid section");
}
if($command eq "apropos") {
    $x = &gettext("apropos", $M, "-r", $page);
    @$x = sort(@$x);
    &render("apropos $page", $x);
} elsif($command eq "whatis") {
    $x = &gettext("whatis", $M, $page);
    @$x = sort(@$x);
    &render("whatis $page", $x);
} elsif($command eq "man") {
    $x = defined $section ? &gettext("man", $M, "-S", $section, $page)
	: &gettext("man", $M, $page);
    &render("man $page", $x);
} else {
    &fail("Unknown command ``$command''");
}

sub fail {
    my $msg = shift;
    print $q->header("text/html");
    print $q->start_html(-title=>$msg);
    $q->param("command", "");
    my $me = $q->self_url;
    print "$msg - <a href=\"$me\">try again</a><p>\n";
    print $q->end_html;
    exit 0;
}

sub gettext {
    my $x = join(" ", @_);
    print STDERR "gettext($x)\n";
    my @x = `$x 2>/dev/null`;
    if($?) {
	&fail("Not found");
    }
    return \@x;
}

sub render {
    my $title = shift;
    my $x = shift;
    my $out;
    local $_;
    print $q->header("text/html");
    print $q->start_html(-title=>$title);
    print "<center><h1>$title</h1></center><hr>\n";
    print "<pre>\n";
    $_ = join("", @$x);
    s/&/\&amp;/g;
    s/</\&lt;/g;
    s/>/\&gt;/g;
    s/_\x08([^&]|&[^;]+;)(\x08\1)+/<b><i>$1<\/i><\/b>/g;
    s/([^&]|&[^;]+;)(\x08\1)+/<b>$1<\/b>/g;
    s/([^&]|&[^;]+;)(\x08_)+/<i>$1<\/i>/g;
    s/_\x08([^&]|&[^;]+;)/<i>$1<\/i>/g;
    s/<\/b><b>//g;
    s/<\/i><i>//g;
    while(/(<[bi]>)?([a-zA-Z0-9_\-\.\+]+([a-zA-Z0-9_\-\.\+:]|(<\/[bi]>)?\n\s+(<[bi]>)?)+)\s*(<\/[bi]>\s*)?(\(\d\w*\))/i) {
	my ($before, $ref, $after) = ($`, $&, $');
	my $page_name = $2;
	$page_name =~ s/(<\/[bi]>)?\n\s+(<[bi]>)?//g;
	my $page_section = $7;
	$page_section = "" if(!defined $page_section);
	my $p = "$page_name$page_section";
	$q->param("command", "man");
	$q->param("page", $p);
	my $me = $q->self_url;
	print $before;
	print "<a href=\"$me\">$ref</a>";
	$_ = $after;
    }
    print $_;
    print "</pre><hr>\n";
    $q->delete("command");
    $q->delete("page");
    my $me = $q->self_url;
    print "<a href=\"$me\">Back to top form...</a>\n";
    print $q->end_html;
}
