#!/bin/sh
#
# INSTALLATION SCRIPT FOR psDooM
#

# Set up default 'root' directory for the install.
DESTDIR=/usr/local

# Set up default install directories.
DBIN=${DESTDIR}/bin
DLIB=${DESTDIR}/games/psdoom
LLIB=${DESTDIR}/lib
DMAN=${DESTDIR}/man

# Print instructions for the user.
cat << EOF


**********************************************************************

                    #########                      ###         ###
                    ##       ##                    ## ##     ## ##
                    ##        ##                   ##  ##   ##  ##
                    ##         ##                  ##   ## ##   ##
    ######  ######  ##         ##  ######  ######  ##    ###    ##
    #    #  #       ##         ##  #    #  #    #  ##           ##
    #    #  ######  ##        ##   #    #  #    #  ##           ##
    #    #       #  ##       ##    #    #  #    #  ##           ##
    ######  ######  #########      ######  ######  ##           ##
    #
    #         Welcome to the psDooM binaries install script!
    #  This script will install the Linux-x86 version of psDooM for
    #  the X Window System and svgalib (console) onto your computer.

When asking for input, the script will show the default choice
in brackets after the prompt.  You may type a different value, or
just press [enter] to accept the default choice.  For yes/no
questions, enter "y" for yes or "n" for no.

**********************************************************************

EOF

echo "Ready to install psDooM? [n]"

read DOIT

if [ "X${DOIT}" != "Xy" -a "X${DOIT}" != "XY" -a "X${DOIT}" != "Xyes" -a "X${DOIT}" != "XYES" ]; then
   echo ${DOIT}
   echo "psDooM installation aborted."
   exit
fi

cat << EOF


**********************************************************************

You may specify the root directory where you want to install psDooM.

The default root directory is:
${DESTDIR}

Files will be installed in the folowing directories:
${DBIN}
${DLIB}
${LLIB}/acc/xdoom
${DMAN}/man6

Note that LLIB (set here as [${LLIB}])
and LLIBDIR (set in the xwadtools Makefile) must be the same,
otherwise acc won't find its include files.

**********************************************************************

EOF

echo "Enter a root directory name. [${DESTDIR}]"

read NEWDESTDIR

if [ ! -z "${NEWDESTDIR}" ]; then
   DESTDIR=${NEWDESTDIR}
fi

DBIN=${DESTDIR}/bin
DLIB=${DESTDIR}/games/psdoom
LLIB=${DESTDIR}/lib
DMAN=${DESTDIR}/man

cat << EOF


**********************************************************************

The root directory for the psDooM install is:
${DESTDIR}

psDooM files will be installed in the folowing directories:
${DBIN}
${DLIB}
${LLIB}/acc/xdoom
${DMAN}/man6

**********************************************************************

EOF

echo "Complete psDooM install with these directories? [n]"

read DOIT

if [ "X${DOIT}" != "Xy" -a "X${DOIT}" != "XY" -a "X${DOIT}" != "Xyes" -a "X${DOIT}" != "XYES" ]; then
   echo ${DOIT}
   echo "psDooM installation aborted."
   exit
fi


# Create the ps-sxdoom.sh and ps-xdoom.sh scripts on the fly to
# incorporate changes to the install path.

# ps-sxdoom.sh
cat << EOF > ps-sxdoom.sh
#!/bin/sh
#
LDIR=${DLIB}

[ "\${DOOMWADDIR}" ] || DOOMWADDIR=\${LDIR}
PATH=\${LDIR}:\${PATH}
export DOOMWADDIR PATH

exec \${LDIR}/ps-sxdoom \$*
EOF

# ps-xdoom.sh
cat << EOF > ps-xdoom.sh
#!/bin/sh
#
LDIR=${DLIB}

[ "\${DOOMWADDIR}" ] || DOOMWADDIR=\${LDIR}
PATH=\${LDIR}:\${PATH}
export DOOMWADDIR PATH

exec \${LDIR}/ps-xdoom \$*
EOF


# Create destination directories, copy the files, and set permissions.
mkdir -p ${DBIN}
mkdir -p ${DLIB}
mkdir -p ${LLIB}/acc/xdoom
mkdir -p ${DMAN}/man6
cp musserver.bin ${DLIB}
cp musserver ${DLIB}
cp sndserver ${DLIB}
cp ps-xdoom ${DLIB}
cp ps-sxdoom ${DLIB}
cp ps-xdoom.sh ${DBIN}/ps-xdoom
cp ps-sxdoom.sh ${DBIN}/ps-sxdoom
chmod 755 ${DLIB}/musserver.bin
chmod 755 ${DLIB}/musserver
chmod 755 ${DLIB}/sndserver
chmod 755 ${DLIB}/ps-xdoom
chmod 755 ${DLIB}/ps-sxdoom
chmod 755 ${DBIN}/ps-xdoom
chmod 755 ${DBIN}/ps-sxdoom
cp xdoom.6 ${DMAN}/man6
cp xdoom.wad ${DLIB}
cp common.acs ${LLIB}/acc/xdoom
cp specials.acs ${LLIB}/acc/xdoom
cp defs.acs ${LLIB}/acc/xdoom

# Let user know install is complete.
cat << EOF


**********************************************************************

                    #########                      ###         ###
                    ##       ##                    ## ##     ## ##
                    ##        ##                   ##  ##   ##  ##
                    ##         ##                  ##   ## ##   ##
    ######  ######  ##         ##  ######  ######  ##    ###    ##
    #    #  #       ##         ##  #    #  #    #  ##           ##
    #    #  ######  ##        ##   #    #  #    #  ##           ##
    #    #       #  ##       ##    #    #  #    #  ##           ##
    ######  ######  #########      ######  ######  ##           ##
    #
    #                    INSTALL COMPLETED!
    #      Run 'ps-xdoom' for the X Windows System version.
    #     Run 'ps-sxdoom' for the svgalib (console) version.

Be sure to put an IWAD (doom1.wad, doom.wad, doomu.wad, doom2.wad) in
${DLIB}, or the game will fail with an
'Error: Game mode indeterminate' message.

Have FUN!!!

**********************************************************************

EOF
